import enum


class EnumCheckResult(enum.IntFlag):
    SKIP = 0
    CAPTURE = 1
    STOP = 2
    CAPTURE_AND_STOP = 3


class WindowStyle(enum.IntFlag):
    """The object that holds the window styles.

    Ref: https://docs.microsoft.com/en-us/windows/win32/winmsg/window-styles
    """

    BORDER = 0x00800000
    CAPTION = 0x00C00000
    CHILD = 0x40000000
    CHILDWINDOW = 0x40000000
    CLIPCHILDREN = 0x02000000
    CLIPSIBLINGS = 0x04000000
    DISABLED = 0x08000000
    DLGFRAME = 0x00400000
    GROUP = 0x00020000
    HSCROLL = 0x00100000
    ICONIC = 0x20000000
    MAXIMIZE = 0x01000000
    MAXIMIZEBOX = 0x00010000
    MINIMIZE = 0x20000000
    MINIMIZEBOX = 0x00020000
    OVERLAPPED = 0x00000000
    POPUP = 0x80000000
    SIZEBOX = 0x00040000
    SYSMENU = 0x00080000
    TABSTOP = 0x00010000
    THICKFRAME = 0x00040000
    TILED = 0x00000000
    TOOLWINDOW = 0x00000080
    VISIBLE = 0x10000000
    VSCROLL = 0x00200000
    OVERLAPPEDWINDOW = (
        OVERLAPPED | CAPTION | SYSMENU | THICKFRAME | MINIMIZEBOX | MAXIMIZEBOX
    )
    POPUPWINDOW = POPUP | BORDER | SYSMENU
    TILEDWINDOW = (
        OVERLAPPED | CAPTION | SYSMENU | THICKFRAME | MINIMIZEBOX | MAXIMIZEBOX
    )


class WindowExStyle(enum.IntFlag):
    """The object that holds the extended window styles.

    Ref: https://docs.microsoft.com/en-us/windows/win32/winmsg/extended-window-styles>
    """

    ACCEPTFILES = 0x00000010
    APPWINDOW = 0x00040000
    CLIENTEDGE = 0x00000200
    COMPOSITED = 0x02000000
    CONTEXTHELP = 0x00000400
    CONTROLPARENT = 0x00010000
    DLGMODALFRAME = 0x00000001
    LAYERED = 0x00080000
    LAYOUTRTL = 0x00400000
    LEFT = 0x00000000
    LEFTSCROLLBAR = 0x00004000
    LTRREADING = 0x00000000
    MDICHILD = 0x00000040
    NOACTIVATE = 0x08000000
    NOINHERITLAYOUT = 0x00100000
    NOPARENTNOTIFY = 0x00000004
    NOREDIRECTIONBITMAP = 0x00200000
    RIGHT = 0x00001000
    RIGHTSCROLLBAR = 0x00000000
    RTLREADING = 0x00002000
    STATICEDGE = 0x00020000
    TOOLWINDOW = 0x00000080
    TOPMOST = 0x00000008
    TRANSPARENT = 0x00000020
    WINDOWEDGE = 0x00000100
    OVERLAPPEDWINDOW = WINDOWEDGE | CLIENTEDGE
    PALETTEWINDOW = WINDOWEDGE | TOOLWINDOW | TOPMOST


class ShowWindowCmd(enum.IntFlag):
    """The object that holds the CmdShow for ShowWindow api

    Ref: https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-showwindow
    """

    SW_HIDE = 0
    SW_MAXIMIZE = 3
    SW_MINIMIZE = 6
    SW_RESTORE = 9
    SW_SHOW = 5
    SW_SHOWMAXIMIZED = 3
    SW_SHOWMINIMIZED = 2
    SW_SHOWMINNOACTIVE = 7
    SW_SHOWNA = 8
    SW_SHOWNOACTIVATE = 4
    SW_SHOWNORMAL = 1


class DwmWindowAttribute(enum.IntEnum):
    """Options used by the DwmGetWindowAttribute and DwmSetWindowAttribute functions.

    Ref: https://learn.microsoft.com/en-us/windows/win32/api/dwmapi/ne-dwmapi-dwmwindowattribute
    """

    DWMWA_NCRENDERING_ENABLED = 1
    DWMWA_NCRENDERING_POLICY = 2
    DWMWA_TRANSITIONS_FORCEDISABLED = 3
    DWMWA_ALLOW_NCPAINT = 4
    DWMWA_CAPTION_BUTTON_BOUNDS = 5
    DWMWA_NONCLIENT_RTL_LAYOUT = 6
    DWMWA_FORCE_ICONIC_REPRESENTATION = 7
    DWMWA_FLIP3D_POLICY = 8
    DWMWA_EXTENDED_FRAME_BOUNDS = 9
    DWMWA_HAS_ICONIC_BITMAP = 10
    DWMWA_DISALLOW_PEEK = 11
    DWMWA_EXCLUDED_FROM_PEEK = 12
    DWMWA_CLOAK = 13
    DWMWA_CLOAKED = 14
    DWMWA_FREEZE_REPRESENTATION = 15
    DWMWA_PASSIVE_UPDATE_MODE = 16
    DWMWA_USE_HOSTBACKDROPBRUSH = 17
    DWMWA_USE_IMMERSIVE_DARK_MODE = (20,)
    DWMWA_WINDOW_CORNER_PREFERENCE = (33,)
    DWMWA_BORDER_COLOR = 34
    DWMWA_CAPTION_COLOR = 35
    DWMWA_TEXT_COLOR = 36
    DWMWA_VISIBLE_FRAME_BORDER_THICKNESS = 37
    DWMWA_SYSTEMBACKDROP_TYPE = 38
    DWMWA_LAST = 39



from jigsawwm.jmk.hotkey import *
from jigsawwm.w32.vk import *


def test_hotkey(mocker):
    hotkey = JmkHotkeys(mocker.Mock())
    win_alt_a_cb = mocker.Mock()
    hotkey.register([Vk.LWIN, Vk.LMENU, Vk.A], win_alt_a_cb)
    # trigger when key up
    hotkey(JmkEvent(Vk.LWIN, True))
    hotkey(JmkEvent(Vk.LMENU, True))
    hotkey(JmkEvent(Vk.A, True))
    assert win_alt_a_cb.call_count == 0
    hotkey(JmkEvent(Vk.A, False))
    time.sleep(0.1)
    assert win_alt_a_cb.call_count == 1
    # second trigger
    hotkey(JmkEvent(Vk.A, True))
    assert win_alt_a_cb.call_count == 1
    hotkey(JmkEvent(Vk.A, False))
    time.sleep(0.01)
    assert win_alt_a_cb.call_count == 2
    assert hotkey.next_handler.call_count == 2
    # pass through if modifier key is released first
    hotkey(JmkEvent(Vk.A, True))
    hotkey(JmkEvent(Vk.LMENU, False))
    hotkey(JmkEvent(Vk.A, False))
    hotkey(JmkEvent(Vk.LWIN, False))
    assert win_alt_a_cb.call_count == 2
    assert hotkey.next_handler.call_count == 6



import logging
from dataclasses import dataclass
from typing import Callable, List, Tuple, Union

from jigsawwm.w32.vk import Vk, parse_combination

from .core import *

logger = logging.getLogger(__name__)

JmkHotkeyComb = Union[List[Vk], str]


@dataclass
class JmkHotkey:
    keys: typing.List[Vk]
    callback: typing.Callable


class JmkHotkeys(JmkHandler):
    """A handler that handles hotkeys."""

    next_handler: JmkHandler
    combs: typing.Dict[typing.FrozenSet[Vk], JmkHotkey]
    pressed_modifiers: typing.Set[Vk]
    resend: JmkEvent

    def __init__(
        self, next_handler, hotkeys: List[Tuple[JmkHotkeyComb, Callable]] = None
    ):
        self.next_handler = next_handler
        self.combs = {}
        self.pressed_modifiers = set()
        self.resend = None
        if hotkeys:
            for comb, cb in hotkeys:
                self.register(comb, cb)

    @staticmethod
    def expand_comb(comb: JmkHotkeyComb) -> List[List[Vk]]:
        if isinstance(comb, str):
            comb = parse_combination(comb)
        for key in comb[:-1]:
            if key not in Modifers:
                raise TypeError("hotkey keys must be a list of Modifers and a Vk")
            if comb[-1] in Modifers:
                raise TypeError("hotkey keys must be a list of Modifers and a Vk")
        return expand_combination(comb)

    def register(self, comb: JmkHotkeyComb, cb: Union[Callable, str]):
        if isinstance(cb, str):
            new_comb = parse_combination(cb)
            cb = lambda: send_combination(*new_comb)
        for keys in self.expand_comb(comb):
            hotkey = JmkHotkey(keys, cb)
            self.combs[frozenset(keys)] = hotkey

    def unregister(self, comb: JmkHotkeyComb):
        for keys in self.expand_comb(comb):
            self.combs.pop(frozenset(keys))

    def find_hotkey(self, evt: JmkEvent) -> typing.Optional[JmkHotkey]:
        pressed_keys = self.pressed_modifiers.copy()
        pressed_keys.add(evt.vk)
        hotkey = self.combs.get(frozenset(pressed_keys))
        # wheel up/down don't have pressed event
        if evt.vk == Vk.WHEEL_UP or evt.vk == Vk.WHEEL_DOWN:
            pressed_keys.add(evt.vk)
        logger.debug("current pressed keys: %s", pressed_keys)
        return hotkey

    def __call__(self, evt: JmkEvent) -> bool:
        logger.debug("%s >>> hotkey", evt)
        if evt.pressed:
            if evt.vk in Modifers:
                self.pressed_modifiers.add(evt.vk)
            else:
                # swallow non-modifier keypress event if hotkey is registered
                hotkey = self.find_hotkey(evt)
                if hotkey and hotkey.keys[-1] == evt.vk:
                    evt.system = False
                    self.resend = evt
                    return True
        else:
            if evt.vk in self.pressed_modifiers:
                self.pressed_modifiers.remove(evt.vk)
            else:
                hotkey = self.find_hotkey(evt)
                if hotkey:
                    if len(hotkey.keys) == 2 and hotkey.keys[0] in (
                        Vk.LWIN,
                        Vk.RWIN,
                    ):
                        # prevent start menu from popping up
                        self.next_handler(JmkEvent(Vk.NONAME, False))
                    logger.info("hotkey triggered: %s", hotkey.keys)
                    execute(hotkey.callback)
                    return True  # maybe let user define whether to swallow
                elif (
                    self.resend
                ):  # modifier key released first, so we resend previous event
                    self.next_handler(self.resend)
                    self.resend = None

        return self.next_handler(evt)
